/*
  universe.cpp

  Copyright (c) 2020 Kevin Matz (kevin.matz@gmail.com)

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in all
  copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.
*/

#include "universe.h"

namespace DMX {

/**
 * @brief Universe::Universe
 * @param timeout_period
 */
Universe::Universe(int timeout_period)
  : rx_timeout_period_(timeout_period)
{
  null_start_data_.fill(0);
}


/**
 * @brief Universe::~Universe
 */
Universe::~Universe()
{
}


/**
 * @brief Universe::data
 * @return
 */
const DimmerData * Universe::data() const
{
  return &null_start_data_;
}


/**
 * @brief Universe::slot
 * @param address
 * @return
 */
uint8_t Universe::slot(const uint16_t address) const
{
  if (address == 0)
    return 0;
  if (address > null_start_data_.size() - 1)
    return 0;
  return null_start_data_[address];
}


/**
 * @brief Universe::rxRate Rate (in Hz) at which the universe is being recv'd.
 * @return Hz
 *
 *  Calculated as the rolling mean of (the number of frames in
 *  the last rx_timeout_peroid ms) / rx_timeout_period seconds.
 */
double Universe::rxRate()
{
    rx_timeout_();
    //         updates     per      second
    return rx_times_.size() / (rx_timeout_period_ / 1000.0);
}


/**
 * @brief Universe::setData Accept new data from receiver
 * @param vect varaibly sized data, beginning with a start code.
 *
 * The only setData method used to establish rx rate.
 */
void Universe::setData(std::vector<uint8_t> vect) {
  switch (vect.front()) // start code
    {
    case E111_NULL_START:
      vect.resize(null_start_data_.size(), 0); // pad shorter, truncate larger
      std::copy(vect.begin(), vect.end(), null_start_data_.begin());
      rx_timeout_(true);
      for (const auto &cb : callbacks_)
        cb(this);
      break;
    default:
      break;
    }
}


/**
 * @brief Universe::setValue
 * @param address
 * @param value
 */
void Universe::setValue(const uint16_t address, const uint8_t value)
{
    if (address == 0)
      return;
    if (address > null_start_data_.size() - 1)
      return;
    null_start_data_[address] = value;
}


/**
 * @brief Universe::setValue
 * @param start
 * @param footprint
 * @param profile
 */
void Universe::setValue(const uint16_t start, const uint16_t footprint,
                   const uint8_t* profile)
{
    if (start == 0)
      return;
    if (start + footprint > null_start_data_.size() - 1)
      return;
    for (int i = 0; i < footprint; i++)
        null_start_data_[start + i] = profile[i];
}


/**
 * @brief Universe::onData
 * @param callback
 */
void Universe::onData(const DataHandler callback)
{
  callbacks_.push_back(callback);
}


/**
 * @brief Universe::rx_timeout_
 * @param add_now
 */
void Universe::rx_timeout_(bool add_now)
{
  auto now = std::chrono::system_clock::now();
  auto elapsed = [] (auto& a, auto& b) {
      return std::chrono::duration_cast<std::chrono::milliseconds>(a - b);
    };

  if (add_now)
    rx_times_.push(now);

  while (rx_times_.size() > 0 &&
         elapsed(now, rx_times_.front()).count() > rx_timeout_period_)
    rx_times_.pop();
}

} // namespace DMX
