/*
  rlp/component.h

  Copyright (c) 2021 Kevin Matz (kevin.matz@gmail.com)

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in all
  copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.
*/

#include "component.h"
#include "tcp.h"
#include "udp.h"

namespace ACN::RLP {

/**
 * @brief Component::Component
 * @param cid
 */
Component::Component(UUID::uuid cid)
  : ACN::Component(cid)
{
  fctn_ = "OpenLCP RLP Component";
};


/**
 * @brief Component::UdpPayloadReceiver
 * @param stream
 *
 * EPI 17
 */
void Component::UdpPayloadReceiver(PDU::Stream stream)
{
  auto transport = RLP::UDP::transport();
  transport.iStream(stream);

  for(auto const &root : *transport.root.pdu) {
    RlpReceiver(root);
  }
}


/**
 * @brief Component::rlpSendUdp
 * @param vector
 * @param data
 * @param ip
 */
void Component::rlpSendUdp (const uint32_t vector,
                            const PDU::Message< PDU::pdu_data > data,
                            const SDT::UDP::ipAddress& ip) const
{
  RLP::UDP::transport transport(true);
  rlpSend(transport, vector, data, ip);
}


/**
 * @brief Component::TcpPacketReceiver
 * @param stream
 *
 * EPI 33
 */
void Component::TcpPacketReceiver(PDU::Stream stream)
{
  auto transport = RLP::TCP::transport();
  transport.iStream(stream);

  for(auto const &root : *transport.root.pdu) {
    RlpReceiver(root);
  }
}


/**
 * @brief Component::rlpSendTcp
 * @param vector
 * @param data
 * @param ip
 */
void Component::rlpSendTcp (const uint32_t vector,
                            const PDU::Message<PDU::pdu_data> data,
                            const SDT::UDP::ipAddress& ip) const
{
  RLP::TCP::transport transport(true);
  rlpSend(transport, vector, data, ip);
}


/**
 * @brief Component::RlpReceiver
 * @param root
 *
 * Dispatch a recieved RLP PDU to the appropriate vector handlers.
 */
void Component::RlpReceiver(PDU::Message<RLP::Pdu> root)
{
  /// To this point, the data segment of the RLP PDU will not have been read.
  /// The root PDU must have not failed, and still have bytes available to read.
  if (!root->stream()->good())
    return;

  if (rlp_vectors_.count(root->vector()))
    for(auto const &handler : rlp_vectors_[root->vector()])
      handler(root);
}


/**
 * @brief Component::RlpRegisterVectorHandler
 * @param vect
 * @param handle
 *
 * Add callback handler for a given RLP vector.
 */
void Component::RlpRegisterVector(uint32_t vect, PDU::Handler<RLP::Pdu> handle)
{
  rlp_vectors_[vect].push_back(handle);
}


/**
 * @brief Component::rlpSend
 * @param transport
 * @param vector
 * @param data
 * @param ip
 */
void Component::rlpSend(ACN::RLP::message_transport& transport,
                        const uint32_t vector,
                        const PDU::Message<PDU::pdu_data> data,
                        const SDT::UDP::ipAddress& ip) const
{
  // rlp header
  RLP::rlp_header header;
  header.cid = cid();

  // rlp pdu
  auto rlp = std::shared_ptr<RLP::Pdu>(new RLP::Pdu());
  rlp->setVector(vector);
  rlp->setHeader(&header);
  rlp->setData(data.get());

  // root block
  PDU::Block<RLP::Pdu> block;
  block.pdu->push_back(rlp);
  transport.root = block;

  // create an output stream
  size_t length = transport.streamSize();
  uint8_t buffer[length];
  ACN::PDU::Stream stream(new ACN::PDU::pdu_stream(buffer, length));

  // write message
  *stream << transport;
  rlpSend(stream, ip);
}


/**
 * @brief Component::rlpSend
 * @param stream
 * @param ip
 *
 * \attention Override this method in your derived class that can write
 * to the network.
 */
void Component::rlpSend(const PDU::Stream stream,
                        const SDT::UDP::ipAddress& ip) const
{
  (void)stream;
  (void)ip;
}


}; // ACN::RLP
