/*
  sender.h

  Copyright (c) 2021 Kevin Matz (kevin.matz@gmail.com)

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in all
  copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.
*/
#pragma once

#include "rlp/component.h"
#include "sacn.h"
#include "universe.h"

#include <unordered_map>

namespace sACN {

/**
 * @brief \cite sACN 3.5 Source: A stream of E1.31 Packets for a universe is
 * said to be sent from a source.
 *
 * > A source is uniquely identified by a number in the header of the packet
 * > (see field CID in Table 4-1). A source may output multiple streams of data,
 * > each for a different universe. Also, multiple sources may output data for
 * > a given universe.
 */
class Source
    : public virtual ACN::RLP::Component
{
public:
    Source(UUID::uuid = UUID::uuid());
    virtual void create(const uint16_t);
    virtual void terminate(const uint16_t);
    Universe * universe(const uint16_t);

protected:
    virtual void end(const uint16_t);

private:
    std::unordered_map <uint16_t, Universe *> universes_;

    /// > \cite sACN 6.2.5 E1.31 Data Packet: Sequence Number
    /// >
    /// > Sources shall maintain a sequence for each universe they transmit.
    /// > The sequence number for a universe shall be incremented by one for
    /// > every packet sent on that universe. There is no implied relationship
    /// > between the sequence number of an E1.31 Synchronization Packet and
    /// > the sequence number of an E1.31 Data Packet on that same universe.
    std::unordered_map <uint16_t, uint8_t> data_sequences_;
    std::unordered_map <uint16_t, uint8_t> sync_sequences_;
};


} // SACN
